# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import unittest
import random
import wget
from gen_vm_image.utils.io import join, find, remove, exists, makedirs
from gen_vm_image.architecture import load_architecture
from gen_vm_image.cli.build_image import create_image, convert_image, resize_image


class TestImageBuild(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # https://cloud.debian.org/images/cloud/bookworm/latest/input_path_image-genericcloud-amd64.qcow2
        # Download the test image for the path input test
        cls.input_path_url = "https://cloud.debian.org/images/cloud/bookworm/latest/debian-12-genericcloud-amd64.qcow2"
        cls.test_tmp_directory = join("tests", "tmp")
        cls.test_res_directory = join("tests", "res")
        if not exists(cls.test_tmp_directory):
            assert makedirs(cls.test_tmp_directory)
        cls.input_image_path = join(cls.test_tmp_directory, "debian-12.qcow2")
        wget.download(cls.input_path_url, out=cls.input_image_path)
        assert exists(cls.input_image_path)

        cls.architecture_path = join(cls.test_res_directory, "architecture.yml")
        cls.architecture, error = load_architecture(cls.architecture_path)
        assert cls.architecture is not None
        assert error is None
        assert isinstance(cls.architecture, dict)

    @classmethod
    def tearDownClass(cls):
        # Remove the downloaded test image
        if exists(cls.input_image_path):
            assert remove(cls.input_image_path)[0]
        if exists(cls.test_tmp_directory):
            assert remove(cls.test_tmp_directory, recursive=True)[0]

    def setUp(self):
        self.seed = str(random.random())[2:10]

    def tearDown(self):
        # Remove every seed file
        regex_name = ".*{}.*".format(self.seed)
        seed_files = find(self.test_tmp_directory, regex_name)
        for seed_file in seed_files:
            if exists(seed_file):
                self.assertTrue(remove(seed_file)[0])

    def test_create_image_qcow_1(self):
        image_1 = self.architecture["images"]["image-1"]
        self.assertEqual(image_1["name"], "test-image-1")
        image_1_name = "{}-{}".format(image_1["name"], self.seed)
        self.assertEqual(image_1["version"], 9.4)
        self.assertEqual(image_1["format"], "qcow2")
        self.assertEqual(image_1["size"], "10G")

        new_image_path = join(
            self.test_tmp_directory, "{}.{}".format(image_1_name, image_1["format"])
        )
        result, msg = create_image(
            new_image_path,
            image_1["size"],
            image_format=image_1["format"],
        )
        self.assertTrue(result)
        self.assertIsNone(msg)

    def test_create_image_raw_2(self):
        image_2 = self.architecture["images"]["image-2"]
        self.assertEqual(image_2["name"], "test-image-2")
        image_2_name = "{}-{}".format(image_2["name"], self.seed)
        self.assertEqual(image_2["version"], 9.4)
        self.assertEqual(image_2["format"], "raw")
        self.assertEqual(image_2["size"], "10G")

        new_image_path = join(
            self.test_tmp_directory, "{}.{}".format(image_2_name, image_2["format"])
        )
        result, msg = create_image(
            new_image_path,
            image_2["size"],
            image_format=image_2["format"],
        )
        self.assertTrue(result)
        self.assertIsNone(msg)

    def test_image_path_input(self):
        input_path_image = self.architecture["images"]["input_path_image"]
        self.assertEqual(input_path_image["name"], "input-path-image")
        self.assertEqual(input_path_image["version"], 12)
        self.assertEqual(input_path_image["format"], "qcow2")
        self.assertEqual(input_path_image["size"], "20G")

        self.assertIn("input", input_path_image)
        self.assertIsInstance(input_path_image["input"], dict)
        self.assertIn("path", input_path_image["input"])
        self.assertIn("format", input_path_image["input"])

        self.assertEqual(input_path_image["input"]["path"], self.input_image_path)
        self.assertEqual(input_path_image["input"]["format"], "qcow2")
        self.assertTrue(exists(self.input_image_path))

        output_path = join(
            self.test_tmp_directory,
            "{}-{}.{}".format(
                input_path_image["name"], self.seed, input_path_image["format"]
            ),
        )

        concert_result, convert_msg = convert_image(
            input_path_image["input"]["path"],
            output_path,
            input_format=input_path_image["input"]["format"],
            output_format=input_path_image["format"],
        )
        self.assertTrue(concert_result)
        self.assertIsNone(convert_msg)
        self.assertTrue(exists(output_path))

        resize_result, resize_msg = resize_image(
            output_path,
            input_path_image["size"],
            image_format=input_path_image["format"],
        )
        self.assertTrue(resize_result)
        self.assertIsNone(resize_msg)
        # TODO, validate that the size of the output image is correct

    def test_convert_image_format(self):
        convert_image_dict = self.architecture["images"]["convert_input_image_format"]
        self.assertEqual(convert_image_dict["name"], "convert_input_image_format")
        self.assertEqual(convert_image_dict["version"], 12)
        self.assertEqual(convert_image_dict["format"], "raw")
        self.assertEqual(convert_image_dict["size"], "20G")

        self.assertIn("input", convert_image_dict)
        self.assertGreater(len(convert_image_dict["input"]), 0)
        self.assertIn("path", convert_image_dict["input"])
        self.assertIn("format", convert_image_dict["input"])
        self.assertEqual(convert_image_dict["input"]["format"], "qcow2")

        self.assertTrue(exists(convert_image_dict["input"]["path"]))
        output_path = join(
            self.test_tmp_directory,
            "{}-{}.{}".format(
                convert_image_dict["name"], self.seed, convert_image_dict["format"]
            ),
        )

        result, msg = convert_image(
            convert_image_dict["input"]["path"],
            output_path,
            input_format=convert_image_dict["input"]["format"],
            output_format=convert_image_dict["format"],
        )
        self.assertTrue(result)
        self.assertIsNone(msg)
        self.assertTrue(exists(output_path))
        # TODO, validate that the output image is of the correct size and format

    def test_build_with_no_version(self):
        image_no_version = self.architecture["images"]["non_version_image_test"]
        self.assertEqual(image_no_version["name"], "non-version-image")
        self.assertEqual(image_no_version["format"], "raw")
        self.assertEqual(image_no_version["size"], "10G")

        new_image_path = join(
            self.test_tmp_directory,
            "{}-{}.{}".format(
                image_no_version["name"], self.seed, image_no_version["format"]
            ),
        )
        result, msg = create_image(
            new_image_path,
            image_no_version["size"],
            image_format=image_no_version["format"],
        )
        self.assertTrue(result)
        self.assertIsNone(msg)
        self.assertTrue(exists(new_image_path))



# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import argparse
import os
import yaml
from gen_vm_image._version import __version__
from gen_vm_image.common.defaults import (
    GOCD_GROUP,
    GOCD_TEMPLATE,
    REPO_NAME,
    PACKAGE_NAME,
    GENERATED_IMAGE_DIR,
    TMP_DIR,
    GOCD_FORMAT_VERSION,
    GO_REVISION_COMMIT_VAR,
    CONSITENCY_SUPPPORTED_FORMATS,
)
from gen_vm_image.common.errors import (
    PATH_NOT_FOUND_ERROR,
    PATH_NOT_FOUND_ERROR_MSG,
    PATH_CREATE_ERROR,
    PATH_CREATE_ERROR_MSG,
    MISSING_ATTRIBUTE_ERROR,
    MISSING_ATTRIBUTE_ERROR_MSG,
    INVALID_ATTRIBUTE_TYPE_ERROR,
    INVALID_ATTRIBUTE_TYPE_ERROR_MSG,
    CHECKSUM_ERROR,
    RESIZE_ERROR,
    RESIZE_ERROR_MSG,
    CHECK_ERROR,
    CHECK_ERROR_MSG,
)
from gen_vm_image.architecture import load_architecture, correct_architecture_structure
from gen_vm_image.utils.io import exists, makedirs, write, hashsum
from gen_vm_image.utils.job import run
from gen_vm_image.utils.net import download_file


current_dir = os.path.abspath(os.path.dirname(__file__))
parent_dir = os.path.dirname(current_dir)


def get_pipelines(images):
    pipelines = []
    for build in images:
        pipelines.append(build)
    return pipelines


def get_common_environment(pipelines):
    common_environment = {
        "environments": {
            GOCD_GROUP: {
                "environment_variables": {
                    "GIT_USER": "{{SECRET:[github][username]}}",
                },
                "pipelines": pipelines,
            }
        }
    }
    return common_environment


def get_common_pipeline():
    common_pipeline = {
        "group": GOCD_GROUP,
        "label_template": "${COUNT}",
        "lock_behaviour": "none",
        "display_order": -1,
        "template": GOCD_TEMPLATE,
    }
    return common_pipeline


def get_common_materials(branch="main"):
    common_materials = {
        "ucphhpc_images": {
            "git": "https://github.com/ucphhpc/{}.git".format(REPO_NAME),
            "branch": branch,
            "destination": REPO_NAME,
            "username": "${GIT_USER}",
            "password": "{{SECRET:[github][access_token]}}",
        },
        # this is the name of material
        # says about type of material and url at once
    }
    return common_materials


def get_upstream_materials(name, pipeline, stage):
    upstream_materials = {
        "upstream_{}".format(name): {
            "pipeline": pipeline,
            "stage": stage,
        }
    }
    return upstream_materials


def get_materials(name, upstream_pipeline=None, stage=None, branch="main"):
    materials = {}
    common_materials = get_common_materials(branch=branch)
    materials.update(common_materials)
    if upstream_pipeline and stage:
        upstream_materials = get_upstream_materials(name, upstream_pipeline, stage)
        materials.update(upstream_materials)
    return materials


def qemu_img_call(action, args, format_output_str=True, verbose=False):
    command = ["qemu-img", action]
    if not verbose:
        command.append("-q")
    command.extend(args)

    result = run(command, format_output_str=format_output_str)
    if result["returncode"] != "0":
        return False, result["error"]
    return True, None


def create_image(path, size, image_format="qcow2", verbose=False):
    args = ["-f", image_format, path, size]
    result, msg = qemu_img_call("create", args, verbose=verbose)
    if not result:
        return PATH_CREATE_ERROR, PATH_CREATE_ERROR_MSG.format(path, msg)
    return result, None


def convert_image(
    input_path, output_path, input_format="qcow2", output_format="qcow2", verbose=False
):
    args = ["-f", input_format, "-O", output_format, input_path, output_path]
    result, msg = qemu_img_call("convert", args, verbose=verbose)
    if not result:
        return PATH_CREATE_ERROR, PATH_CREATE_ERROR_MSG.format(input_path, msg)
    return result, None


def resize_image(path, size, image_format="qcow2", verbose=False):
    result, msg = qemu_img_call(
        "resize", ["-f", image_format, path, size], verbose=verbose
    )
    if not result:
        return RESIZE_ERROR, RESIZE_ERROR_MSG.format(msg)
    return result, None


def amend_image(path, options, image_format="qcow2", verbose=False):
    args = ["-f", image_format, "-o", options, path]
    result, msg = qemu_img_call("amend", args, verbose=verbose)
    if not result:
        return PATH_CREATE_ERROR, PATH_CREATE_ERROR_MSG.format(path, msg)
    return result, None


def check_image(path, image_format="qcow2", verbose=False):
    if image_format not in CONSITENCY_SUPPPORTED_FORMATS:
        msg = "format: '{}' not supported for consistency check, only one of: {} is supported".format(
            image_format, CONSITENCY_SUPPPORTED_FORMATS
        )
        return False, msg
    result, msg = qemu_img_call("check", ["-f", image_format, path], verbose=verbose)
    if not result:
        return CHECK_ERROR, CHECK_ERROR_MSG.format(msg)
    return result, None


def build_gocd_config(architecture, gocd_name, branch, verbose=False):
    images = architecture.get("images", [])

    # GOCD file
    list_images = list(images.keys())

    # Get all pipelines
    pipelines = get_pipelines(list_images)

    # GOCD environment
    common_environments = get_common_environment(pipelines)

    # Common GOCD pipeline params
    common_pipeline_attributes = get_common_pipeline()

    generated_config = {
        "format_version": GOCD_FORMAT_VERSION,
        **common_environments,
        "pipelines": {},
    }
    # Generate the GOCD build config
    for build, build_data in images.items():
        name = build_data.get("name", None)
        version = build_data.get("version", None)
        materials = get_materials(name, branch=branch)

        build_version_name = build
        build_pipeline = {
            **common_pipeline_attributes,
            "materials": materials,
            "parameters": {
                "IMAGE": name,
                "IMAGE_PIPELINE": build_version_name,
                "DEFAULT_TAG": version,
                "SRC_DIRECTORY": REPO_NAME,
                "TEST_DIRECTORY": REPO_NAME,
                "PUSH_DIRECTORY": "publish-python-scripts",
                "COMMIT_TAG": GO_REVISION_COMMIT_VAR,
                "ARGS": "",
            },
        }
        generated_config["pipelines"][build_version_name] = build_pipeline

    gocd_config_path = os.path.join(current_dir, gocd_name)
    if not write(gocd_config_path, generated_config, handler=yaml):
        print(
            PATH_CREATE_ERROR_MSG.format(gocd_config_path, "Failed to save gocd config")
        )
        exit(PATH_CREATE_ERROR)
    if verbose:
        print("Generated a new GOCD config in: {}".format(gocd_config_path))


def build_architecture(
    architecture_path, images_output_directory, overwrite=False, verbose=False
):
    # Load the architecture file
    architecture, load_error_msg = load_architecture(architecture_path)
    if load_error_msg:
        print(load_error_msg)
        exit(architecture)

    correct_architecture, correct_error_msg = correct_architecture_structure(
        architecture
    )
    if not correct_architecture:
        print(correct_error_msg)
        exit(correct_architecture)

    # Create the destination directory where the images will be saved
    if not exists(images_output_directory):
        created, msg = makedirs(images_output_directory)
        if not created:
            print(PATH_CREATE_ERROR_MSG.format(images_output_directory, msg))
            exit(PATH_CREATE_ERROR)

    if "images" not in architecture:
        print(MISSING_ATTRIBUTE_ERROR_MSG.format("images", architecture))
        exit(MISSING_ATTRIBUTE_ERROR)

    images = architecture.get("images")
    if not isinstance(images, dict):
        print(INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(type(images), images, dict))
        exit(INVALID_ATTRIBUTE_TYPE_ERROR)

    # Generate the image configuration
    for build, build_data in images.items():
        vm_name = build_data["name"]
        vm_size = build_data["size"]
        # Optional version attribute for each image configuration
        vm_version = build_data.get("version", None)

        # Optional attributes for each image configuration
        vm_output_format = build_data.get("format", "qcow2")
        vm_input, vm_input_format = None, "qcow2"
        if "input" in build_data:
            vm_input = build_data["input"]
            if not isinstance(vm_input, str) and not isinstance(vm_input, dict):
                print(
                    INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                        type(vm_input), vm_input, "string or dictionary"
                    )
                )
                exit(INVALID_ATTRIBUTE_TYPE_ERROR)

        if vm_version:
            vm_output_path = os.path.join(
                images_output_directory,
                "{}-{}.{}".format(vm_name, vm_version, vm_output_format),
            )
        else:
            vm_output_path = os.path.join(
                images_output_directory, "{}.{}".format(vm_name, vm_output_format)
            )

        if exists(vm_output_path):
            if verbose:
                print("The output image: {} already exists".format(vm_output_path))
            if not overwrite:
                if verbose:
                    print(
                        "Use the --overwrite flag to overwrite the existing image, continuing to the next image..."
                    )
                continue
            else:
                if verbose:
                    print("Overwriting the existing image: {}".format(vm_output_path))

        if vm_input:
            if isinstance(vm_input, dict):
                if "url" not in vm_input and "path" not in vm_input:
                    print(
                        MISSING_ATTRIBUTE_ERROR_MSG.format(
                            "'url' or 'path' must be in the architecture input section",
                            vm_input,
                        )
                    )
                    exit(MISSING_ATTRIBUTE_ERROR)

                if "url" in vm_input and "path" in vm_input:
                    print(
                        "Both 'url' and 'path' are defined in the architecture input section. Only one can be defined"
                    )
                    exit(INVALID_ATTRIBUTE_TYPE_ERROR)

                if "url" in vm_input:
                    if not isinstance(vm_input["url"], str):
                        print(
                            INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                                type(vm_input["url"]), vm_input["url"], "string"
                            )
                        )
                        exit(INVALID_ATTRIBUTE_TYPE_ERROR)

                if "path" in vm_input:
                    if not isinstance(vm_input["path"], str):
                        print(
                            INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                                type(vm_input["path"]), vm_input["path"], "string"
                            )
                        )
                        exit(INVALID_ATTRIBUTE_TYPE_ERROR)

                # If a checksum is present, then validate that it is correctly structured
                vm_input_checksum = None
                if "checksum" in vm_input:
                    if not isinstance(vm_input["checksum"], dict):
                        print(
                            INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                                type(vm_input["checksum"]),
                                vm_input["checksum"],
                                "dictionary",
                            )
                        )
                        exit(INVALID_ATTRIBUTE_TYPE_ERROR)

                    required_checksum_attributes = ["type", "value"]
                    for attr in required_checksum_attributes:
                        if attr not in vm_input["checksum"]:
                            print(
                                MISSING_ATTRIBUTE_ERROR_MSG.format(
                                    attr, vm_input["checksum"]
                                )
                            )
                            exit(MISSING_ATTRIBUTE_ERROR)
                        if not isinstance(vm_input["checksum"][attr], str):
                            print(
                                INVALID_ATTRIBUTE_TYPE_ERROR_MSG.format(
                                    type(vm_input["checksum"][attr]),
                                    vm_input["checksum"][attr],
                                    "string",
                                )
                            )
                            exit(INVALID_ATTRIBUTE_TYPE_ERROR)
                    vm_input_checksum = vm_input["checksum"]

                if "url" in vm_input:
                    vm_input_url = vm_input["url"]
                    # Download the specified url and save it into
                    # a tmp directory.
                    # First prepare the temporary directory
                    # where the downloaded image will be prepared
                    if not exists(TMP_DIR):
                        created, msg = makedirs(TMP_DIR)
                        if not created:
                            print(PATH_CREATE_ERROR_MSG.format(TMP_DIR, msg))
                            exit(PATH_CREATE_ERROR)

                    input_url_filename = vm_input_url.split("/")[-1]
                    input_vm_path = os.path.join(TMP_DIR, input_url_filename)
                    if not exists(input_vm_path):
                        if verbose:
                            print("Downloading image from: {}".format(vm_input_url))
                        download_file(vm_input_url, input_vm_path)
                elif "path" in vm_input:
                    input_vm_path = vm_input["path"]
                    if not exists(input_vm_path):
                        print(
                            PATH_NOT_FOUND_ERROR_MSG.format(
                                input_vm_path,
                                "the defined input path to the does not exist",
                            )
                        )
                        exit(PATH_NOT_FOUND_ERROR)

                if vm_input_checksum:
                    checksum_type = vm_input_checksum["type"]
                    checksum_value = vm_input_checksum["value"]
                    calculated_checksum = hashsum(
                        input_vm_path, algorithm=checksum_type
                    )
                    if not calculated_checksum:
                        print(
                            "Failed to calculate the checksum of the downloaded image"
                        )
                        exit(CHECKSUM_ERROR)

                    if calculated_checksum != checksum_value:
                        print(
                            "The checksum of the downloaded image: {} does not match the expected checksum: {}".format(
                                calculated_checksum, checksum_value
                            )
                        )
                        exit(CHECKSUM_ERROR)
                    if verbose:
                        print(
                            "The calculated checksum: {} matches the defined checksum: {}".format(
                                calculated_checksum, checksum_value
                            )
                        )
            else:
                if not exists(vm_input):
                    print(
                        PATH_NOT_FOUND_ERROR_MSG.format(
                            vm_input, "the defined input path to the does not exist"
                        )
                    )
                    exit(PATH_NOT_FOUND_ERROR)

            converted_result, msg = convert_image(
                input_vm_path,
                vm_output_path,
                input_format=vm_input_format,
                output_format=vm_output_format,
                verbose=verbose,
            )
            if not converted_result:
                print(msg)
                exit(converted_result)

            # Resize the vm disk image
            resized_result, resized_msg = resize_image(
                vm_output_path, vm_size, image_format=vm_output_format, verbose=verbose
            )
            if not resized_result:
                print(resized_msg)
                exit(resized_result)
        else:
            # If no input is specified, then we assume that we are creating a new disc image
            create_image_result, msg = create_image(
                vm_output_path,
                vm_size,
                image_format=vm_output_format,
                verbose=verbose,
            )
            if not create_image_result:
                print(msg)
                exit(create_image_result)
            else:
                if verbose:
                    print(
                        "Generated image at: {}".format(os.path.abspath(vm_output_path))
                    )

        # Amend to qcow2 version 3 which is required in RHEL 9 if the output format is
        # qcow2
        # TODO, validate that the image is a rhel based image
        if vm_output_format == "qcow2":
            amend_result, amend_msg = amend_image(
                vm_output_path, "compat=v3", verbose=verbose
            )
            if not amend_result:
                print(PATH_CREATE_ERROR_MSG.format(vm_output_path, amend_msg))

        if vm_output_format in CONSITENCY_SUPPPORTED_FORMATS:
            check_result, check_msg = check_image(
                vm_output_path,
                image_format=vm_output_format,
                verbose=verbose,
            )
            if not check_result:
                print(check_msg)
                exit(check_result)


def cli():
    parser = argparse.ArgumentParser(
        prog=PACKAGE_NAME, formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "architecture_path",
        help="The path to the architecture file that defines the images to build",
    )
    parser.add_argument(
        "--images-output-directory",
        default=GENERATED_IMAGE_DIR,
        help="The path to the output directory where the images will be saved",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        default=False,
        help="Whether the tool should overwrite existing image disks",
    )
    parser.add_argument(
        "--generate-gocd-config",
        action="store_true",
        help="Generate a GoCD config based on the architecture file",
    )
    parser.add_argument(
        "--gocd-config-name",
        default="1.gocd.yml",
        help="Name of the generated gocd config file",
    )
    parser.add_argument(
        "--gocd-build-branch",
        default="main",
        help="The branch that GoCD should use to build images",
    )
    parser.add_argument(
        "--verbose",
        "-v",
        action="store_true",
        default=False,
        help="Print verbose output",
    )
    parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print the version of the program",
    )
    args = parser.parse_args()

    architecture_path = args.architecture_path
    images_output_directory = args.images_output_directory
    overwrite = args.overwrite
    generate_gocd_config = args.generate_gocd_config
    gocd_config_name = args.gocd_config_name
    gocd_build_branch = args.gocd_build_branch
    verbose = args.verbose

    build_architecture(
        architecture_path, images_output_directory, overwrite=overwrite, verbose=verbose
    )
    if generate_gocd_config:
        build_gocd_config(
            architecture_path, gocd_config_name, gocd_build_branch, verbose=verbose
        )
    exit(0)


if __name__ == "__main__":
    cli()



# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import requests
import tqdm


def download_file(url, ouput_path, chunk_size=8192):
    with open(ouput_path, "wb") as _file:
        with requests.get(url, stream=True) as r:
            r.raise_for_status()
            total = int(r.headers.get("content-length", 0))

            tqdm_params = {
                "desc": url,
                "total": total,
                "miniters": 1,
                "unit": "B",
                "unit_scale": True,
                "unit_divisor": 1024,
            }
            with tqdm.tqdm(**tqdm_params) as pb:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    pb.update(len(chunk))
                    _file.write(chunk)



# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import yaml
from gen_vm_image.common.errors import (
    PATH_NOT_FOUND_ERROR,
    PATH_NOT_FOUND_ERROR_MSG,
    PATH_LOAD_ERROR,
    PATH_LOAD_ERROR_MSG,
    MISSING_ATTRIBUTE_ERROR,
    MISSING_ATTRIBUTE_ERROR_MSG,
)
from gen_vm_image.utils.io import exists, load


def load_architecture(architecture_path):
    if not exists(architecture_path):
        return PATH_NOT_FOUND_ERROR, PATH_NOT_FOUND_ERROR_MSG.format(
            architecture_path, "Failed to find the architecture file."
        )
    architecture = load(architecture_path, handler=yaml, Loader=yaml.FullLoader)
    if not architecture:
        return PATH_LOAD_ERROR, PATH_LOAD_ERROR_MSG.format(
            architecture_path, "Failed to load the architecture file."
        )
    return architecture, None


def correct_architecture_structure(architecture):
    owner = architecture.get("owner", None)
    if not owner:
        return MISSING_ATTRIBUTE_ERROR, MISSING_ATTRIBUTE_ERROR_MSG.format(
            "owner", architecture
        )
    images = architecture.get("images", None)
    if not images:
        return MISSING_ATTRIBUTE_ERROR, MISSING_ATTRIBUTE_ERROR_MSG.format(
            "images", architecture
        )

    required_attributes = ["name", "version", "size"]
    for image_name, image_data in images.items():
        for attribute in required_attributes:
            if attribute not in image_data:
                return MISSING_ATTRIBUTE_ERROR, MISSING_ATTRIBUTE_ERROR_MSG.format(
                    attribute, image_name
                )
    return True, None



# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import unittest
from gen_vm_image.utils.io import join
from gen_vm_image.architecture import load_architecture
from gen_vm_image.common.errors import PATH_NOT_FOUND_ERROR


class TestImageArchitecture(unittest.TestCase):
    def setUp(self):
        self.architecture_path = join("tests", "res", "architecture.yml")

    def test_load_architecture(self):
        architecture, error = load_architecture(self.architecture_path)
        self.assertNotEqual(architecture, None)
        self.assertEqual(error, None)
        self.assertIsInstance(architecture, dict)

    def test_fail_load_architecture(self):
        missing_architecture_path = join("tests", "res", "missing.yml")
        architecture, error = load_architecture(missing_architecture_path)
        self.assertNotEqual(error, None)
        self.assertEqual(architecture, PATH_NOT_FOUND_ERROR)

    def test_architecture_owner(self):
        architecture, error = load_architecture(self.architecture_path)
        self.assertNotEqual(architecture, None)
        self.assertEqual(error, None)
        self.assertIsInstance(architecture, dict)
        self.assertIn("owner", architecture)
        self.assertIsInstance(architecture["owner"], str)
        self.assertGreater(len(architecture["owner"]), 0)
        self.assertEqual(architecture["owner"], "the-owner-name")

    def test_architecture_images_top(self):
        architecture, error = load_architecture(self.architecture_path)
        self.assertNotEqual(architecture, None)
        self.assertEqual(error, None)
        self.assertIsInstance(architecture, dict)
        self.assertIn("images", architecture)
        self.assertIsInstance(architecture["images"], dict)
        self.assertGreater(len(architecture["images"]), 0)

    def test_architecture_images_items(self):
        architecture, error = load_architecture(self.architecture_path)
        self.assertNotEqual(architecture, None)
        self.assertEqual(error, None)
        self.assertIsInstance(architecture, dict)
        self.assertIn("images", architecture)
        self.assertIsInstance(architecture["images"], dict)
        self.assertGreater(len(architecture["images"]), 0)
        self.assertIn("image-1", architecture["images"])
        images = architecture["images"]

        for image_name, image_data in images.items():
            self.assertIsInstance(image_name, str)
            self.assertNotEqual(image_name, None)
            self.assertNotEqual(image_name, "")

            self.assertIsInstance(image_data, dict)
            self.assertIn("name", image_data)
            self.assertIsInstance(image_data["name"], str)
            self.assertGreater(len(image_data["name"]), 0)

            # Optional version attribute
            if "version" in image_data:
                self.assertIsInstance(image_data["version"], (float, int))

            self.assertIn("format", image_data)
            self.assertIsInstance(image_data["format"], str)
            self.assertGreater(len(image_data["format"]), 0)
            self.assertIn(image_data["format"], ["raw", "qcow2"])

            self.assertIn("size", image_data)
            self.assertIsInstance(image_data["size"], str)
            self.assertGreater(len(image_data["size"]), 0)



# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import pwd
import grp


def lookup_uid(username):
    try:
        return pwd.getpwnam(username).pw_uid
    except Exception as err:
        print("Failed to discover user uid: {} - {}".format(username, err))
    return False


def lookup_gid(groupname):
    try:
        return grp.getgrnam(groupname).gr_gid
    except Exception as err:
        print("Failed to discover group gid: {} - {}".format(groupname, err))
    return False


def find_user_with_username(username):
    for user in pwd.getpwall():
        if username in user.pw_name:
            return user.pw_name
    return False


def find_group_with_groupname(groupname):
    for group in grp.getgrall():
        if groupname in group.gr_name:
            return group.gr_name
    return False



# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import subprocess


def __format_output__(result, format_output_str=False):
    command_results = {}
    if hasattr(result, "args"):
        command_results.update({"command": " ".join((getattr(result, "args")))})
    if hasattr(result, "returncode"):
        command_results.update({"returncode": getattr(result, "returncode")})
    if hasattr(result, "stderr"):
        command_results.update({"error": getattr(result, "stderr")})
    if hasattr(result, "stdout"):
        command_results.update({"output": getattr(result, "stdout")})
    if hasattr(result, "wait"):
        command_results.update({"wait": getattr(result, "wait")})
    if hasattr(result, "communicate"):
        command_results.update({"communicate": getattr(result, "communicate")})
    if hasattr(result, "kill"):
        command_results.update({"kill": getattr(result, "kill")})

    if format_output_str:
        for key, value in command_results.items():
            if key == "returncode" or key == "stderr" or key == "stdout":
                command_results[key] = str(value)
    return command_results


def run_popen(cmd, format_output_str=False, **run_kwargs):
    result = subprocess.Popen(cmd, **run_kwargs)
    return __format_output__(result, format_output_str=format_output_str)


def check_call(cmd, format_output_str=False, **kwargs):
    result = subprocess.check_call(cmd, **kwargs)
    return __format_output__(result, format_output_str=format_output_str)


def run(cmd, format_output_str=False, **run_kwargs):
    result = subprocess.run(cmd, **run_kwargs)
    return __format_output__(result, format_output_str=format_output_str)



# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

PATH_NOT_FOUND_ERROR = -1
PATH_NOT_FOUND_ERROR_MSG = "Path not found: {} - error: {}"
PATH_LOAD_ERROR = -2
PATH_LOAD_ERROR_MSG = "Failed to load path: {}"
PATH_CREATE_ERROR = -3
PATH_CREATE_ERROR_MSG = "Failed to create path: {} - error: {}"
MISSING_ATTRIBUTE_ERROR = -4
MISSING_ATTRIBUTE_ERROR_MSG = "Missing attribute: {} in {}"
INVALID_ATTRIBUTE_TYPE_ERROR = -5
INVALID_ATTRIBUTE_TYPE_ERROR_MSG = (
    "Invalid attribute type: {} for value: {} - must be {}"
)
CHECKSUM_ERROR = -6
RESIZE_ERROR = -7
RESIZE_ERROR_MSG = "Failed to resize path: {}"
CHECK_ERROR = -8
CHECK_ERROR_MSG = "Failed to check path: {}"



#!/usr/bin/python
# coding: utf-8
# Copyright (C) 2024  The gen-vm-image Project by the Science HPC Center at UCPH
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
from setuptools import setup, find_packages

cur_dir = os.path.abspath(os.path.dirname(__file__))


def read(path):
    with open(path, "r") as _file:
        return _file.read()


def read_req(name):
    path = os.path.join(cur_dir, name)
    return [req.strip() for req in read(path).splitlines() if req.strip()]


version_ns = {}
version_path = os.path.join(cur_dir, "gen_vm_image", "_version.py")
version_content = read(version_path)
exec(version_content, {}, version_ns)


long_description = open("README.rst").read()
setup(
    name="gen-vm-image",
    version=version_ns["__version__"],
    description="This tool can be used for generating virtual machine images.",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Rasmus Munk",
    author_email="munk1@live.dk",
    packages=find_packages(),
    url="https://github.com/ucphhpc/gen-vm-image",
    license="GNU General Public License v2 (GPLv2)",
    keywords=["Virtual Machine", "VM", "Images"],
    install_requires=read_req("requirements.txt"),
    extras_require={
        "test": read_req("tests/requirements.txt"),
        "dev": read_req("requirements-dev.txt"),
    },
    entry_points={
        "console_scripts": [
            "gen-vm-image = gen_vm_image.cli.build_image:cli",
        ]
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
    ],
)
